/**
 * @file exifiron.c
 * Orientation correction and optimization of Exif JPEG images
 * @author Marko Mkel (marko.makela@iki.fi)
 */

/* Copyright  2003,2004,2005 Marko Mkel.

   This file is part of EXIFIRON, a program for orientation correction and
   shrinking of digital photographs in JPEG EXIF format.

   EXIFIRON is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   EXIFIRON is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

/*
 * The code for lossless JPEG image transformations (rotating and flipping)
 * has been derived from the file transupp.c, which is part of the jpegtran
 * utility of the Independent JPEG Group,
 * Copyright  1997-2001 Thomas G. Lane.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * According to comments in transupp.c, the transformations were initially
 * designed and coded by Guido Vollbeding.
 *
 * In this file, we have removed the options for cropping and trimming.
 */

#if defined WIN32 || defined __WIN32
# undef __STRICT_ANSI__
# include <sys/types.h>
# include <sys/utime.h>
#else
# ifndef _POSIX_SOURCE
#  define _POSIX_SOURCE
# endif
# include <sys/types.h>
# include <utime.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <time.h>

#include "rename.h"

/* Although this file really should not have access to the library internals,
 * it is helpful to let it call jround_up () and jcopy_block_row ().
 */
#define JPEG_INTERNALS

#include "jpeglib.h"
#include "jerror.h"

/* code adapted from transupp.c begins */

/** Lossless image transformations (according to the Orientation tag) */
enum JXFORM_CODE {
  JXFORM_NONE = 1,	/**< no transformation */
  JXFORM_FLIP_H,	/**< horizontal flip */
  JXFORM_ROT_180,	/**< 180-degree rotation */
  JXFORM_FLIP_V,	/**< vertical flip */
  JXFORM_TRANSPOSE,	/**< transpose across UL-to-LR axis */
  JXFORM_ROT_90,	/**< 90-degree clockwise rotation */
  JXFORM_TRANSVERSE,	/**< transpose across UR-to-LL axis */
  JXFORM_ROT_270	/**< 270-degree clockwise (or 90 ccw) */
};

/** JPEG image transformation information.
 * This structure has been adapted from transupp.c.
 */
struct jpeg_transform_info
{
  /** @name Options: set by caller */
  /*@{*/
  /** image transform operator */
  enum JXFORM_CODE transform;
  /*@}*/

  /** Internal workspace: caller should not touch these */
  /*@{*/
  /** workspace for transformations */
  struct jvirt_barray_control** workspace_coef_arrays;
  /** destination width */
  JDIMENSION output_width;
  /** destination height */
  JDIMENSION output_height;
  /** destination sampling factor (iMCU size), horizontal */
  int max_h_samp_factor;
  /** destination sampling factor (iMCU size), vertical */
  int max_v_samp_factor;
  /*@}*/
};

/** Horizontal flip; done in-place, so no separate dest array is required.
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 */
static void
flip_h (struct jpeg_decompress_struct* srcinfo,
	struct jpeg_compress_struct* dstinfo,
	struct jvirt_barray_control** src_coef_arrays)
{
  int ci;

  /* Horizontal mirroring of DCT blocks is accomplished by swapping
   * pairs of blocks in-place.  Within a DCT block, we perform horizontal
   * mirroring by changing the signs of odd-numbered columns.
   * Partial iMCUs at the right edge are left untouched.
   */
  const JDIMENSION MCU_cols =
    srcinfo->image_width / (dstinfo->max_h_samp_factor * DCTSIZE);

  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    const JDIMENSION comp_width = MCU_cols * compptr->h_samp_factor;
    JDIMENSION blk_y, blk_x;
    for (blk_y = 0; blk_y < compptr->height_in_blocks;
	 blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, src_coef_arrays[ci], blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE);
      int offset_y;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	/* Do the mirroring */
	for (blk_x = 0; blk_x * 2 < comp_width; blk_x++) {
	  JCOEFPTR ptr1 = buffer[offset_y][blk_x];
	  JCOEFPTR ptr2 = buffer[offset_y][comp_width - blk_x - 1];
	  /* this unrolled loop doesn't need to know which row it's on... */
	  int k = DCTSIZE2 / 2;
	  while (k--) {
	    /* swap even column */
	    JCOEF temp1 = *ptr1, temp2 = *ptr2;
	    *ptr1++ = temp2, *ptr2++ = temp1;
	    /* swap odd column with sign change */
	    temp1 = *ptr1, temp2 = *ptr2;
	    *ptr1++ = -temp2, *ptr2++ = -temp1;
	  }
	}
      }
    }
  }
}

/** Vertical flip.
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param dst_coef_arrays	coefficient arrays for the destination image
 */
static void
flip_v (struct jpeg_decompress_struct* srcinfo,
	struct jpeg_compress_struct* dstinfo,
	struct jvirt_barray_control** src_coef_arrays,
	struct jvirt_barray_control** dst_coef_arrays)
{
  int ci;
  
  /* We output into a separate array because we can't touch different
   * rows of the source virtual array simultaneously.  Otherwise, this
   * is a pretty straightforward analog of horizontal flip.
   * Within a DCT block, vertical mirroring is done by changing the signs
   * of odd-numbered rows.
   * Partial iMCUs at the bottom edge are copied verbatim.
   */
  const JDIMENSION MCU_rows =
    srcinfo->image_height / (dstinfo->max_v_samp_factor * DCTSIZE);

  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    const JDIMENSION comp_height = MCU_rows * compptr->v_samp_factor;
    JDIMENSION dst_blk_y, dst_blk_x;
    for (dst_blk_y = 0; dst_blk_y < compptr->height_in_blocks;
	 dst_blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const dst_buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, dst_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE);
      JBLOCKARRAY const src_buffer = dst_blk_y < comp_height
	? (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, src_coef_arrays[ci],
	 comp_height - dst_blk_y -
	 (JDIMENSION) compptr->v_samp_factor,
	 (JDIMENSION) compptr->v_samp_factor, FALSE)
	: (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, src_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, FALSE);
      int offset_y;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	if (dst_blk_y < comp_height) {
	  /* Row is within the mirrorable area. */
	  JBLOCKROW const dst_row_ptr = dst_buffer[offset_y];
	  JBLOCKROW const src_row_ptr =
	    src_buffer[compptr->v_samp_factor - offset_y - 1];
	  for (dst_blk_x = 0; dst_blk_x < compptr->width_in_blocks;
	       dst_blk_x++) {
	    JCOEFPTR
	      dst_ptr = dst_row_ptr[dst_blk_x],
	      src_ptr = src_row_ptr[dst_blk_x];
	    int i = DCTSIZE / 2;
	    while (i--) {
	      int j;
	      /* copy even row */
	      memcpy (dst_ptr, src_ptr, DCTSIZE * sizeof *dst_ptr);
	      dst_ptr += DCTSIZE, src_ptr += DCTSIZE;
	      /* copy odd row with sign change */
	      for (j = 0; j < DCTSIZE; j++)
		*dst_ptr++ = - *src_ptr++;
	    }
	  }
	}
	else
	  jcopy_block_row (src_buffer[offset_y], dst_buffer[offset_y],
			   compptr->width_in_blocks);
      }
    }
  }
}

/** Transpose source into destination.
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param dst_coef_arrays	coefficient arrays for the destination image
 */
static void
transpose (struct jpeg_decompress_struct* srcinfo,
	   struct jpeg_compress_struct* dstinfo,
	   struct jvirt_barray_control** src_coef_arrays,
	   struct jvirt_barray_control** dst_coef_arrays)
{
  int ci;

  /* Transposing pixels within a block just requires transposing the
   * DCT coefficients.
   * Partial iMCUs at the edges require no special treatment; we simply
   * process all the available DCT blocks for every component.
   */
  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    JDIMENSION dst_blk_y, dst_blk_x;
    for (dst_blk_y = 0; dst_blk_y < compptr->height_in_blocks;
	 dst_blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const dst_buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, dst_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE);
      int offset_y, offset_x;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	for (dst_blk_x = 0; dst_blk_x < compptr->width_in_blocks;
	     dst_blk_x += compptr->h_samp_factor) {
	  JBLOCKARRAY const src_buffer = (*srcinfo->mem->access_virt_barray)
	    ((j_common_ptr) srcinfo, src_coef_arrays[ci], dst_blk_x,
	     (JDIMENSION) compptr->h_samp_factor, FALSE);
	  for (offset_x = 0; offset_x < compptr->h_samp_factor; offset_x++) {
	    JCOEFPTR const
	      src_ptr = src_buffer[offset_x][dst_blk_y + offset_y],
	      dst_ptr = dst_buffer[offset_y][dst_blk_x + offset_x];
	    int i, j;
	    for (i = 0; i < DCTSIZE; i++)
	      for (j = 0; j < DCTSIZE; j++)
		dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
	  }
	}
      }
    }
  }
}

/** 90 degree clockwise rotation (transposition and horizontal mirroring).
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param dst_coef_arrays	coefficient arrays for the destination image
 */
static void
rot90 (struct jpeg_decompress_struct* srcinfo,
       struct jpeg_compress_struct* dstinfo,
       struct jvirt_barray_control** src_coef_arrays,
       struct jvirt_barray_control** dst_coef_arrays)
{
  int ci;

  /* Because of the horizontal mirror step, we can't process partial iMCUs
   * at the (output) right edge properly.  They just get transposed and
   * not mirrored.
   */
  const JDIMENSION MCU_cols =
    srcinfo->image_height / (dstinfo->max_h_samp_factor * DCTSIZE);

  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    const JDIMENSION comp_width = MCU_cols * compptr->h_samp_factor;
    JDIMENSION dst_blk_y, dst_blk_x;
    for (dst_blk_y = 0; dst_blk_y < compptr->height_in_blocks;
	 dst_blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const dst_buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, dst_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE);
      int offset_y, offset_x;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	for (dst_blk_x = 0; dst_blk_x < compptr->width_in_blocks;
	     dst_blk_x += compptr->h_samp_factor) {
	  JBLOCKARRAY const src_buffer = (*srcinfo->mem->access_virt_barray)
	    ((j_common_ptr) srcinfo, src_coef_arrays[ci], dst_blk_x,
	     (JDIMENSION) compptr->h_samp_factor, FALSE);
	  for (offset_x = 0; offset_x < compptr->h_samp_factor; offset_x++) {
	    int i, j;
	    JCOEFPTR const src_ptr =
	      src_buffer[offset_x][dst_blk_y + offset_y];
	    if (dst_blk_x < comp_width) {
	      /* Block is within the mirrorable area. */
	      JCOEFPTR const dst_ptr = dst_buffer[offset_y]
		[comp_width - dst_blk_x - offset_x - 1];
	      for (i = 0; i < DCTSIZE; i++) {
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
		i++;
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = -src_ptr[i*DCTSIZE+j];
	      }	      
	    }
	    else {
	      /* Edge blocks are transposed but not mirrored. */
	      JCOEFPTR const dst_ptr =
		dst_buffer[offset_y][dst_blk_x + offset_x];
	      for (i = 0; i < DCTSIZE; i++)
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
	    }
	  }
	}
      }
    }
  }
}

/** 270 degree clockwise rotation (horizontal mirroring and transposition).
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param dst_coef_arrays	coefficient arrays for the destination image
 */
static void
rot270 (struct jpeg_decompress_struct* srcinfo,
	struct jpeg_compress_struct* dstinfo,
	struct jvirt_barray_control** src_coef_arrays,
	struct jvirt_barray_control** dst_coef_arrays)
{
  int ci;

  /* Because of the horizontal mirror step, we can't process partial iMCUs
   * at the (output) bottom edge properly.  They just get transposed and
   * not mirrored.
   */
  const JDIMENSION MCU_rows =
    srcinfo->image_width / (dstinfo->max_v_samp_factor * DCTSIZE);

  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    const JDIMENSION comp_height = MCU_rows * compptr->v_samp_factor;
    JDIMENSION dst_blk_y, dst_blk_x;
    for (dst_blk_y = 0; dst_blk_y < compptr->height_in_blocks;
	 dst_blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const dst_buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, dst_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE);
      int offset_y, offset_x;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	for (dst_blk_x = 0; dst_blk_x < compptr->width_in_blocks;
	     dst_blk_x += compptr->h_samp_factor) {
	  JBLOCKARRAY const src_buffer = (*srcinfo->mem->access_virt_barray)
	    ((j_common_ptr) srcinfo, src_coef_arrays[ci],
	     dst_blk_x,
	     (JDIMENSION) compptr->h_samp_factor, FALSE);
	  for (offset_x = 0; offset_x < compptr->h_samp_factor; offset_x++) {
	    JCOEFPTR const dst_ptr =
	      dst_buffer[offset_y][dst_blk_x + offset_x];
	    int i, j;
	    if (dst_blk_y < comp_height) {
	      /* Block is within the mirrorable area. */
	      JCOEFPTR const src_ptr =
		src_buffer[offset_x][comp_height - dst_blk_y - offset_y - 1];
	      for (i = 0; i < DCTSIZE; i++) {
		for (j = 0; j < DCTSIZE; j++) {
		  dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
		  j++;
		  dst_ptr[j*DCTSIZE+i] = -src_ptr[i*DCTSIZE+j];
		}
	      }
	    }
	    else {
	      /* Edge blocks are transposed but not mirrored. */
	      JCOEFPTR const src_ptr =
		src_buffer[offset_x][dst_blk_y + offset_y];
	      for (i = 0; i < DCTSIZE; i++)
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
	    }
	  }
	}
      }
    }
  }
}

/** 180 degree rotation (vertical and horizontal mirroring).
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param dst_coef_arrays	coefficient arrays for the destination image
 */
static void
rot180 (struct jpeg_decompress_struct* srcinfo,
	struct jpeg_compress_struct* dstinfo,
	struct jvirt_barray_control** src_coef_arrays,
	struct jvirt_barray_control** dst_coef_arrays)
{
  int ci;

  const JDIMENSION
    MCU_cols = srcinfo->image_width / (dstinfo->max_h_samp_factor * DCTSIZE),
    MCU_rows = srcinfo->image_height / (dstinfo->max_v_samp_factor * DCTSIZE);

  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    const JDIMENSION
      comp_width = MCU_cols * compptr->h_samp_factor,
      comp_height = MCU_rows * compptr->v_samp_factor;
    JDIMENSION dst_blk_y, dst_blk_x;
    for (dst_blk_y = 0; dst_blk_y < compptr->height_in_blocks;
	 dst_blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const
	dst_buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, dst_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE),
	src_buffer = dst_blk_y < comp_height
	? (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, src_coef_arrays[ci],
	 comp_height - dst_blk_y -
	 (JDIMENSION) compptr->v_samp_factor,
	 (JDIMENSION) compptr->v_samp_factor, FALSE)
	: (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, src_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, FALSE);
      int offset_y;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	JBLOCKROW const
	  dst_row_ptr = dst_buffer[offset_y],
	  src_row_ptr = dst_blk_y < comp_height
	  ? src_buffer[compptr->v_samp_factor - offset_y - 1]
	  : src_buffer[offset_y];
	JCOEFPTR dst_ptr, src_ptr;
	int i, j;
	if (dst_blk_y < comp_height) {
	  for (dst_blk_x = 0; dst_blk_x < comp_width; dst_blk_x++) {
	    dst_ptr = dst_row_ptr[dst_blk_x];
	    src_ptr = src_row_ptr[comp_width - dst_blk_x - 1];
	    for (i = 0; i < DCTSIZE; i += 2) {
	      /* For even row, negate every odd column. */
	      for (j = 0; j < DCTSIZE; j += 2) {
		*dst_ptr++ = *src_ptr++;
		*dst_ptr++ = - *src_ptr++;
	      }
	      /* For odd row, negate every even column. */
	      for (j = 0; j < DCTSIZE; j += 2) {
		*dst_ptr++ = - *src_ptr++;
		*dst_ptr++ = *src_ptr++;
	      }
	    }
	  }
	  /* Any remaining right-edge blocks are only mirrored vertically. */
	  for (; dst_blk_x < compptr->width_in_blocks; dst_blk_x++) {
	    dst_ptr = dst_row_ptr[dst_blk_x];
	    src_ptr = src_row_ptr[dst_blk_x];
	    for (i = 0; i < DCTSIZE; i += 2) {
	      memcpy (dst_ptr, src_ptr, DCTSIZE * sizeof *dst_ptr);
	      dst_ptr += DCTSIZE, src_ptr += DCTSIZE;
	      for (j = 0; j < DCTSIZE; j++)
		*dst_ptr++ = - *src_ptr++;
	    }
	  }
	}
	else {
	  /* Remaining rows are just mirrored horizontally. */
	  /* Process the blocks that can be mirrored. */
	  for (dst_blk_x = 0; dst_blk_x < comp_width; dst_blk_x++) {
	    dst_ptr = dst_row_ptr[dst_blk_x];
	    src_ptr = src_row_ptr[comp_width - dst_blk_x - 1];
	    for (i = 0; i < DCTSIZE2; i += 2) {
	      *dst_ptr++ = *src_ptr++;
	      *dst_ptr++ = - *src_ptr++;
	    }
	  }
	  /* Any remaining right-edge blocks are only copied. */
	  for (; dst_blk_x < compptr->width_in_blocks; dst_blk_x++)
	    memcpy (dst_row_ptr[dst_blk_x],
		    src_row_ptr[dst_blk_x],
		    DCTSIZE2 * sizeof **dst_row_ptr);
	}
      }
    }
  }
}

/** Transverse (horizontal mirroring, transposition, horizontal mirroring)
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param dst_coef_arrays	coefficient arrays for the destination image
 */
static void
transverse (struct jpeg_decompress_struct* srcinfo,
	    struct jpeg_compress_struct* dstinfo,
	    struct jvirt_barray_control** src_coef_arrays,
	    struct jvirt_barray_control** dst_coef_arrays)
{
  int ci;

  const JDIMENSION
    MCU_cols = srcinfo->image_height / (dstinfo->max_h_samp_factor * DCTSIZE),
    MCU_rows = srcinfo->image_width / (dstinfo->max_v_samp_factor * DCTSIZE);

  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    const JDIMENSION
      comp_width = MCU_cols * compptr->h_samp_factor,
      comp_height = MCU_rows * compptr->v_samp_factor;
    JDIMENSION dst_blk_x, dst_blk_y;
    for (dst_blk_y = 0; dst_blk_y < compptr->height_in_blocks;
	 dst_blk_y += compptr->v_samp_factor) {
      JBLOCKARRAY const dst_buffer = (*srcinfo->mem->access_virt_barray)
	((j_common_ptr) srcinfo, dst_coef_arrays[ci], dst_blk_y,
	 (JDIMENSION) compptr->v_samp_factor, TRUE);
      int offset_y, offset_x;
      for (offset_y = 0; offset_y < compptr->v_samp_factor; offset_y++) {
	for (dst_blk_x = 0; dst_blk_x < compptr->width_in_blocks;
	     dst_blk_x += compptr->h_samp_factor) {
	  JBLOCKARRAY const src_buffer = (*srcinfo->mem->access_virt_barray)
	    ((j_common_ptr) srcinfo, src_coef_arrays[ci], dst_blk_x,
	     (JDIMENSION) compptr->h_samp_factor, FALSE);
	  for (offset_x = 0; offset_x < compptr->h_samp_factor; offset_x++) {
	    JCOEFPTR const
	      dst_ptr = dst_buffer[offset_y]
	      [dst_blk_x < comp_width
	       ? comp_width - dst_blk_x - offset_x - 1
	       : dst_blk_x + offset_x],
	      src_ptr = src_buffer[offset_x]
	      [dst_blk_y < comp_height
	       ? comp_height - dst_blk_y - offset_y - 1
	       : dst_blk_y + offset_y];
	    int i, j;
	    if (dst_blk_y < comp_height) {
	      if (dst_blk_x < comp_width) {
		/* Block is within the mirrorable area. */
		for (i = 0; i < DCTSIZE; i++) {
		  for (j = 0; j < DCTSIZE; j++) {
		    dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
		    j++;
		    dst_ptr[j*DCTSIZE+i] = -src_ptr[i*DCTSIZE+j];
		  }
		  i++;
		  for (j = 0; j < DCTSIZE; j++) {
		    dst_ptr[j*DCTSIZE+i] = -src_ptr[i*DCTSIZE+j];
		    j++;
		    dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
		  }
		}
	      }
	      else {
		/* Right-edge blocks are mirrored in y only */
		for (i = 0; i < DCTSIZE; i++) {
		  for (j = 0; j < DCTSIZE; j++) {
		    dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
		    j++;
		    dst_ptr[j*DCTSIZE+i] = -src_ptr[i*DCTSIZE+j];
		  }
		}
	      }
	    }
	    else if (dst_blk_x < comp_width) {
	      /* Bottom-edge blocks are mirrored in x only */
	      for (i = 0; i < DCTSIZE; i++) {
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
		i++;
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = -src_ptr[i*DCTSIZE+j];
	      }
	    }
	    else {
	      /* At lower right corner, just transpose, no mirroring */
	      for (i = 0; i < DCTSIZE; i++)
		for (j = 0; j < DCTSIZE; j++)
		  dst_ptr[j*DCTSIZE+i] = src_ptr[i*DCTSIZE+j];
	    }
	  }
	}
      }
    }
  }
}

/** Request workspace for the image transformation.
 * This routine figures out the size that the output image will be
 * (which implies that all the transform parameters must be set before
 * it is called).
 *
 * We allocate the workspace virtual arrays from the source decompression
 * object, so that all the arrays (both the original data and the workspace)
 * will be taken into account while making memory management decisions.
 * Hence, this routine must be called after jpeg_read_header (which reads
 * the image dimensions) and before jpeg_read_coefficients (which realizes
 * the source's virtual arrays).

 * The code has been adapted from transupp.c.
 * @param srcinfo	source image information
 * @param info		transformation information
 */
static void
jtransform_request_workspace (struct jpeg_decompress_struct* srcinfo,
			      struct jpeg_transform_info* info)
{
  struct jvirt_barray_control** coef_arrays = 0;
  /* If there is only one output component, force the iMCU size to be 1;
   * else use the source iMCU size.  (This allows us to do the right thing
   * when reducing color to grayscale, and also provides a handy way of
   * cleaning up "funny" grayscale images whose sampling factors are not 11.)
   */

  switch (info->transform) {
  case JXFORM_TRANSPOSE:
  case JXFORM_TRANSVERSE:
  case JXFORM_ROT_90:
  case JXFORM_ROT_270:
    info->output_width = srcinfo->image_height;
    info->output_height = srcinfo->image_width;
    if (srcinfo->num_components == 1) {
      info->max_h_samp_factor = 1;
      info->max_v_samp_factor = 1;
    }
    else {
      info->max_h_samp_factor = srcinfo->max_v_samp_factor;
      info->max_v_samp_factor = srcinfo->max_h_samp_factor;
    }
    break;
  default:
    info->output_width = srcinfo->image_width;
    info->output_height = srcinfo->image_height;
    if (srcinfo->num_components == 1) {
      info->max_h_samp_factor = 1;
      info->max_v_samp_factor = 1;
    }
    else {
      info->max_h_samp_factor = srcinfo->max_h_samp_factor;
      info->max_v_samp_factor = srcinfo->max_v_samp_factor;
    }
    break;
  }

  /* Figure out whether we need workspace arrays,
   * and if so whether they are transposed relative to the source.
   */
  switch (info->transform) {
    int ci;
    JDIMENSION width_in_iMCUs, height_in_iMCUs;
  case JXFORM_NONE:
  case JXFORM_FLIP_H:
    break;
  case JXFORM_ROT_90:
  case JXFORM_ROT_270:
  case JXFORM_TRANSVERSE:
  case JXFORM_TRANSPOSE:
  case JXFORM_ROT_180:
  case JXFORM_FLIP_V:
    /* Allocate workspace.
     * Note that we allocate arrays padded out to the next iMCU boundary,
     * so that transform routines need not worry about missing edge blocks.
     */
    coef_arrays = (struct jvirt_barray_control**)
      (*srcinfo->mem->alloc_small) ((j_common_ptr) srcinfo, JPOOL_IMAGE,
				    sizeof (struct jvirt_barray_control*) *
				    srcinfo->num_components);
    width_in_iMCUs = (JDIMENSION)
      jdiv_round_up((long) info->output_width,
		    (long) (info->max_h_samp_factor * DCTSIZE));
    height_in_iMCUs = (JDIMENSION)
      jdiv_round_up((long) info->output_height,
		    (long) (info->max_v_samp_factor * DCTSIZE));
    for (ci = 0; ci < srcinfo->num_components; ci++) {
      jpeg_component_info* const compptr = srcinfo->comp_info + ci;
      int h_samp_factor, v_samp_factor;
      if (srcinfo->num_components == 1) {
	/* we're going to force samp factors to 11 in this case */
	h_samp_factor = v_samp_factor = 1;
      }
      else if (info->transform == JXFORM_ROT_180 ||
	       info->transform == JXFORM_FLIP_V) {
	h_samp_factor = compptr->h_samp_factor;
	v_samp_factor = compptr->v_samp_factor;
      }
      else { /* transposed */
	h_samp_factor = compptr->v_samp_factor;
	v_samp_factor = compptr->h_samp_factor;
      }
      coef_arrays[ci] = (*srcinfo->mem->request_virt_barray)
	((j_common_ptr) srcinfo, JPOOL_IMAGE, FALSE,
	 width_in_iMCUs * h_samp_factor,
	 height_in_iMCUs * v_samp_factor,
	 (JDIMENSION) v_samp_factor);
    }
  }

  info->workspace_coef_arrays = coef_arrays;
}

/** Transpose destination image parameters.
 * The code has been adapted from transupp.c.
 * @param dstinfo	destination image information
 */
static void
transpose_critical_parameters (struct jpeg_compress_struct* dstinfo)
{
  int tblno, i, j, ci, itemp;
  JQUANT_TBL* qtblptr;
  UINT16 qtemp;

  /* Transpose sampling factors */
  for (ci = 0; ci < dstinfo->num_components; ci++) {
    jpeg_component_info* const compptr = dstinfo->comp_info + ci;
    itemp = compptr->h_samp_factor;
    compptr->h_samp_factor = compptr->v_samp_factor;
    compptr->v_samp_factor = itemp;
  }

  /* Transpose quantization tables */
  for (tblno = 0; tblno < NUM_QUANT_TBLS; tblno++) {
    qtblptr = dstinfo->quant_tbl_ptrs[tblno];
    if (qtblptr != NULL) {
      for (i = 0; i < DCTSIZE; i++) {
	for (j = 0; j < i; j++) {
	  qtemp = qtblptr->quantval[i*DCTSIZE+j];
	  qtblptr->quantval[i*DCTSIZE+j] = qtblptr->quantval[j*DCTSIZE+i];
	  qtblptr->quantval[j*DCTSIZE+i] = qtemp;
	}
      }
    }
  }
}

/** Adjust output image parameters as needed.
 * This must be called after jpeg_copy_critical_parameters ()
 * and before jpeg_write_coefficients ().
 * The code has been adapted from transupp.c.
 * @param dstinfo		destination image information
 * @param src_coef_arrays	coefficient arrays for the source image
 * @param info			transformation information
 * @return	virtual coefficient arrays for jpeg_write_coefficients ()
 */
static struct jvirt_barray_control**
jtransform_adjust_parameters (struct jpeg_compress_struct* dstinfo,
			      struct jvirt_barray_control** src_coef_arrays,
			      struct jpeg_transform_info* info)
{
  /* Correct the destination's image dimensions as necessary
   * for rotate/flip operations.
   */
  dstinfo->image_width = info->output_width;
  dstinfo->image_height = info->output_height;

  /* Transpose destination image parameters */
  switch (info->transform) {
  case JXFORM_TRANSPOSE:
  case JXFORM_TRANSVERSE:
  case JXFORM_ROT_90:
  case JXFORM_ROT_270:
    transpose_critical_parameters (dstinfo);
    /*@fallthrough@*/
  case JXFORM_NONE:
  case JXFORM_FLIP_H:
  case JXFORM_FLIP_V:
  case JXFORM_ROT_180:
    break;
  }

  return info->workspace_coef_arrays
    ? info->workspace_coef_arrays
    : src_coef_arrays;
}

/** Execute the actual transformation, if any.
 * The code has been adapted from transupp.c.
 * @param srcinfo		source image information
 * @param dstinfo		dest. image info from jpeg_write_coefficients
 * @param src_coef_arrays	coefficient arrays for the source image
 * (modified by most transformations)
 * @param info			transformation information
 */
static void
jtransform_execute_transform (struct jpeg_decompress_struct* srcinfo,
			      struct jpeg_compress_struct* dstinfo,
			      struct jvirt_barray_control** src_coef_arrays,
			      struct jpeg_transform_info* info)
{
  struct jvirt_barray_control** dst_coef_arrays = info->workspace_coef_arrays;

  /* Note: conditions tested here should match those in switch statement
   * in jtransform_request_workspace ()
   */
  switch (info->transform) {
  case JXFORM_NONE:
    break;
  case JXFORM_FLIP_H:
    flip_h (srcinfo, dstinfo, src_coef_arrays);
    break;
  case JXFORM_FLIP_V:
    flip_v (srcinfo, dstinfo, src_coef_arrays, dst_coef_arrays);
    break;
  case JXFORM_TRANSPOSE:
    transpose (srcinfo, dstinfo, src_coef_arrays, dst_coef_arrays);
    break;
  case JXFORM_TRANSVERSE:
    transverse (srcinfo, dstinfo, src_coef_arrays, dst_coef_arrays);
    break;
  case JXFORM_ROT_90:
    rot90 (srcinfo, dstinfo, src_coef_arrays, dst_coef_arrays);
    break;
  case JXFORM_ROT_180:
    rot180 (srcinfo, dstinfo, src_coef_arrays, dst_coef_arrays);
    break;
  case JXFORM_ROT_270:
    rot270 (srcinfo, dstinfo, src_coef_arrays, dst_coef_arrays);
    break;
  }
}

/** Setup decompression object to save desired markers in memory.
 * This must be called before jpeg_read_header () to have the desired effect.
 * The code has been adapted from transupp.c.
 * @param cinfo		source image information
 */
static void
jcopy_markers_setup (struct jpeg_decompress_struct* cinfo)
{
  int m;
  /* save COM markers */
  jpeg_save_markers (cinfo, JPEG_COM, 0xFFFF);
  /* Save APP0..APP15 markers */
  for (m = 0; m < 16; m++)
    jpeg_save_markers (cinfo, JPEG_APP0 + m, 0xFFFF);
}

/** Copy markers saved in the given source object to the destination object.
 * This should be called just after jpeg_start_compress () or
 * jpeg_write_coefficients ().
 * Note that those routines will have written the SOI, and also the
 * JFIF APP0 or Adobe APP14 markers if selected.
 * The code has been adapted from transupp.c, and the parameters related to
 * writing thumbnail images have been added.
 * @param srcinfo		source image information
 * @param dstinfo		destination image information
 * @param image			thumbnail image data
 * @param length		length of the thumbnail image data
 */
static void
jcopy_markers_execute (struct jpeg_decompress_struct* srcinfo,
		       struct jpeg_compress_struct* dstinfo,
		       const JOCTET* image,
		       unsigned length)
{
  jpeg_saved_marker_ptr marker;

  /* To avoid confusion, we do not output JFIF and Adobe APP14 markers
   * if the encoder library already wrote one.
   */
  for (marker = srcinfo->marker_list; marker != NULL; marker = marker->next) {
    if (dstinfo->write_JFIF_header &&
	marker->marker == (UINT8) JPEG_APP0 &&
	marker->data_length >= 5 &&
	GETJOCTET (marker->data[0]) == (JOCTET) 0x4A &&
	GETJOCTET (marker->data[1]) == (JOCTET) 0x46 &&
	GETJOCTET (marker->data[2]) == (JOCTET) 0x49 &&
	GETJOCTET (marker->data[3]) == (JOCTET) 0x46 &&
	GETJOCTET (marker->data[4]) == (JOCTET) 0)
      continue;			/* reject duplicate JFIF */
    if (dstinfo->write_Adobe_marker &&
	marker->marker == (UINT8) (JPEG_APP0 + 14) &&
	marker->data_length >= 5 &&
	GETJOCTET (marker->data[0]) == (JOCTET) 0x41 &&
	GETJOCTET (marker->data[1]) == (JOCTET) 0x64 &&
	GETJOCTET (marker->data[2]) == (JOCTET) 0x6F &&
	GETJOCTET (marker->data[3]) == (JOCTET) 0x62 &&
	GETJOCTET (marker->data[4]) == (JOCTET) 0x65)
      continue;			/* reject duplicate Adobe */
    if (length &&
	marker->marker == (UINT8) (JPEG_APP0 + 1) &&
	marker->data_length >= 8 &&
	GETJOCTET (marker->data[0]) == (JOCTET) 0x45 &&
	GETJOCTET (marker->data[1]) == (JOCTET) 0x78 &&
	GETJOCTET (marker->data[2]) == (JOCTET) 0x69 &&
	GETJOCTET (marker->data[3]) == (JOCTET) 0x66 &&
	GETJOCTET (marker->data[4]) == (JOCTET) 0x00 &&
	GETJOCTET (marker->data[5]) == (JOCTET) 0x00) {
      unsigned i;
      /* writing Exif marker with embedded thumbnail */
      jpeg_write_m_header (dstinfo, marker->marker, marker->data_length +
			   length);
      for (i = 0; i < marker->data_length; i++)
	jpeg_write_m_byte (dstinfo, marker->data[i]);
      for (i = 0; i < length; i++)
	jpeg_write_m_byte (dstinfo, image[i]);
      continue;
    }
#ifdef NEED_FAR_POINTERS
    /* We could use jpeg_write_marker if the data weren't FAR... */
    {
      unsigned int i;
      jpeg_write_m_header (dstinfo, marker->marker, marker->data_length);
      for (i = 0; i < marker->data_length; i++)
	jpeg_write_m_byte (dstinfo, marker->data[i]);
    }
#else
    jpeg_write_marker (dstinfo, marker->marker,
		       marker->data, marker->data_length);
#endif
  }
}
/* code adapted from transupp.c ends */

/**
 * Modify the file timestamp
 * @param name		file name
 * @param datetime	modification and access time
 */
static void
touchfile (const char* name,
	   time_t datetime)
{
  struct utimbuf utim;
  utim.actime = utim.modtime = datetime;
  if (utime (name, &utim)) {
    (void) fputs (name, stderr), (void) fflush (stderr);
    perror (": utime");
    /* non-fatal error, don't abort */
  }
}

/** libjpeg error manager */
struct exif_error_mgr
{
  /** "public" fields */
  struct jpeg_error_mgr pub;
  /** for returning to caller */
  jmp_buf setjmp_buffer;
};

/**
 * Custom error exit handler
 * @param cinfo		the libjpeg context
 */
static void
exif_error_exit (struct jpeg_common_struct* cinfo)
{
  longjmp (((struct exif_error_mgr*) cinfo->err)->setjmp_buffer, 1);
}

/** current input file name */
static const char* srcname;

/**
 * Custom message output handler
 * @param cinfo		the libjpeg context
 */
static void
exif_output_message (struct jpeg_common_struct* cinfo)
{
  /** error message buffer for libjpeg */
  static char jpeg_msg[JMSG_LENGTH_MAX];
  cinfo->err->format_message (cinfo, jpeg_msg);
  (void) fputs (srcname, stderr);
  (void) fputs (": ", stderr);
  (void) fputs (jpeg_msg, stderr);
  (void) putc ('\n', stderr);
}

/** error messages from the Exif parser */
static const char* exif_message_table[] = {
  0,
  "Exif: unexpected end of thumbnail image",
  "Exif: transformed thumbnail image is too large",
};

/** error codes from the Exif parser */
enum exif_message {
  EXIF_FIRSTCODE = 1000,	/**< first message code */
  EXIF_THUMBNAIL_EOF,		/**< out of data in Exif thumbnail */
  EXIF_THUMBNAIL_WRITE,		/**< transformed Exif thumbnail too large */
  EXIF_LASTCODE
};

/** dummy method for filling the input buffer
 * @param cinfo	the decompression context
 * @exception	EXIF_THUMBNAIL_EOF
 * @return	TRUE
 */
static boolean
thumb_in_fill (struct jpeg_decompress_struct* cinfo)
{
  ERREXIT (cinfo, EXIF_THUMBNAIL_EOF);
  return TRUE;
}

/** dummy function for image compression into memory buffer
 * @param cinfo	the compression context
 */
static void
thumb_out_dummy (struct jpeg_compress_struct* cinfo
#ifdef __GNUC__
		 __attribute__ ((unused))
#endif /* __GNUC__ */
		 )
{
}

/** Signal that the thumbnail compression buffer is full
 * @param cinfo	the compression context
 * @exception	EXIF_THUMBNAIL_WRITE
 * @return	FALSE, to signal that the data cannot be accepted
 */
static boolean
thumb_out_empty (struct jpeg_compress_struct* cinfo)
{
  ERREXIT (cinfo, EXIF_THUMBNAIL_WRITE);
  return FALSE; /* give up */
}

/** thumbnail compression buffer */
static JOCTET thumb_buf[16384];
/** actual thumbnail length */
static unsigned thumb_length;
/** maximum thumbnail length */
static unsigned thumb_maxlen;

/** transform a thumbnail image in memory
 * @param xform	transformation
 * @param data	image data
 * @param len	source image data length
 * @param maxl	maximum output image data length
 */
static void
thumb_transform (enum JXFORM_CODE xform,
		 JOCTET FAR* data,
		 unsigned len,
		 unsigned maxl)
{
  /** libjpeg decompression context */
  struct jpeg_decompress_struct srcinfo;
  /** libjpeg compression context */
  struct jpeg_compress_struct dstinfo;
  /** libjpeg error handler */
  struct exif_error_mgr jerr;
  struct jpeg_transform_info info;
  struct jvirt_barray_control** src_coef_arrays;
  struct jvirt_barray_control** dst_coef_arrays;

  thumb_maxlen = maxl < (sizeof thumb_buf) / sizeof *thumb_buf
    ? (sizeof thumb_buf) / sizeof *thumb_buf
    : maxl;

  srcinfo.err = dstinfo.err = jpeg_std_error (&jerr.pub);
  jpeg_create_decompress (&srcinfo);
  jpeg_create_compress (&dstinfo);

  jerr.pub.error_exit = exif_error_exit;
  jerr.pub.output_message = exif_output_message;
  jerr.pub.addon_message_table = exif_message_table;
  jerr.pub.first_addon_message = EXIF_FIRSTCODE;
  jerr.pub.last_addon_message = EXIF_LASTCODE;

  if (setjmp (jerr.setjmp_buffer)) {
    exif_output_message ((j_common_ptr) &srcinfo);
    jpeg_abort_decompress (&srcinfo);
    jpeg_abort_compress (&dstinfo);
    goto done;
  }

  jpeg_stdio_src (&srcinfo, 0);
  srcinfo.src->next_input_byte = data;
  srcinfo.src->bytes_in_buffer = len;
  srcinfo.src->fill_input_buffer = thumb_in_fill;
  (void) jpeg_read_header (&srcinfo, TRUE);

  info.transform = xform;
  jtransform_request_workspace (&srcinfo, &info);
  src_coef_arrays = jpeg_read_coefficients (&srcinfo);
  jpeg_copy_critical_parameters (&srcinfo, &dstinfo);
  dst_coef_arrays = jtransform_adjust_parameters (&dstinfo,
						  src_coef_arrays,
						  &info);

  jpeg_stdio_dest (&dstinfo, 0);
  dstinfo.dest->init_destination =
    dstinfo.dest->term_destination = thumb_out_dummy;
  dstinfo.dest->empty_output_buffer = thumb_out_empty;
  dstinfo.dest->next_output_byte = thumb_buf;
  dstinfo.dest->free_in_buffer = thumb_maxlen;
  dstinfo.optimize_coding = TRUE;
  dstinfo.write_JFIF_header = FALSE;
  jpeg_write_coefficients (&dstinfo, dst_coef_arrays);
  jtransform_execute_transform (&srcinfo, &dstinfo, src_coef_arrays, &info);
  (void) jpeg_finish_compress (&dstinfo);
  (void) jpeg_finish_decompress (&srcinfo);
  thumb_length = thumb_maxlen - dstinfo.dest->free_in_buffer;
 done:
  jpeg_destroy_compress (&dstinfo);
  jpeg_destroy_decompress (&srcinfo);
}

/** Byte order of the Exif data */
enum endianness {
  LSB = 0,	/**< least significant byte first (Intel) */
  MSB = 1	/**< most significant byte first (Motorola) */
};

/** Read a 2-octet quantity from the Exif record
 * @param data		offset to the Exif data
 * @param endian	the byte order of the data
 * @return		the quantity read
 */
static unsigned
exif_read2 (const JOCTET FAR* data,
	    enum endianness endian)
{
  return endian
    ? (unsigned) GETJOCTET (data[0]) << 8 | GETJOCTET (data[1])
    : (unsigned) GETJOCTET (data[0]) | GETJOCTET (data[1]) << 8;
}

/** Read a 4-octet quantity from the Exif record
 * @param data		offset to the Exif data
 * @param endian	the byte order of the data
 * @return		the quantity read
 */
static unsigned long
exif_read4 (const JOCTET FAR* data,
	    enum endianness endian)
{
  return endian
    ? ((unsigned long) GETJOCTET (data[0])) << 24 |
    ((unsigned long) GETJOCTET (data[1])) << 16 |
    ((unsigned long) GETJOCTET (data[2])) << 8 |
    ((unsigned long) GETJOCTET (data[3]))
    : ((unsigned long) GETJOCTET (data[0])) |
    ((unsigned long) GETJOCTET (data[1])) << 8 |
    ((unsigned long) GETJOCTET (data[2])) << 16 |
    ((unsigned long) GETJOCTET (data[3])) << 24;
}

/** Write a 2 or 4-octet quantity to the Exif record
 * @param rec		the Exif record
 * @param endian	the byte order of the data
 * @param data		data to be written
 * @return		the quantity read
 */
static void
exif_write (JOCTET FAR* rec,
	    enum endianness endian,
	    unsigned long data)
{
  switch (endian) {
  case LSB:
    rec[8] = (JOCTET) data;
    rec[9] = (JOCTET) (data >> 8);
    rec[10] = (JOCTET) (data >> 16);
    rec[11] = (JOCTET) (data >> 24);
    break;
  case MSB:
    switch (rec[3]) { /* record type */
    case 2: /* string */
    case 4: /* unsigned long */
      rec[11] = (JOCTET) data;
      rec[10] = (JOCTET) (data >> 8);
      rec[9] = (JOCTET) (data >> 16);
      rec[8] = (JOCTET) (data >> 24);
      break;
    case 3: /* unsigned short */
      rec[9] = (JOCTET) data;
      rec[8] = (JOCTET) (data >> 8);
      rec[10] = rec[11] = 0;
      break;
    case 1: case 6: /* byte */
      rec[8] = (JOCTET) data;
      rec[9] = rec[10] = rec[11] = 0;
    }
    break;
  }
}

/** Parse an Exif date stamp and convert it to seconds since the epoch
 * @param date	the date to be checked
 * @return	seconds since the epoch; -1 on failure
 */
static time_t
exif_date (JOCTET FAR* date)
{
  struct tm stamp;
  time_t t;
  int year, month, day, h, m, s;
  unsigned i;
  if (date[19])
    return -1; /* not NUL-terminated */
  for (i = 19; i--; ) {
    switch (i) {
    case 4: case 7: case 13: case 16:
      if (date[i] != (JOCTET) 0x3a/*':'*/)
	return -1; /* improper field terminator */
      break;
    case 10:
      if (date[i] != (JOCTET) 0x20/*' '*/)
	return -1; /* date and time not separated by space */
      break;
    default:
      if (date[i] < (JOCTET) 0x30/*'0'*/ || date[i] > (JOCTET) 0x39/*'9'*/)
	return -1; /* non-digits in date or time */
    }
  }
  year =
    ((int) date[0] - 0x30/*'0'*/) * 1000 +
    ((int) date[1] - 0x30/*'0'*/) * 100 +
    ((int) date[2] - 0x30/*'0'*/) * 10 +
    ((int) date[3] - 0x30/*'0'*/) - 1900;
  month = ((int) date[5] - 0x30/*'0'*/) * 10 +
    ((int) date[6] - 0x30/*'0'*/) - 1;
  day = ((int) date[8] - 0x30/*'0'*/) * 10 + ((int) date[9] - 0x30/*'0'*/);
  h = ((int) date[11] - 0x30/*'0'*/) * 10 + ((int) date[12] - 0x30/*'0'*/);
  m = ((int) date[14] - 0x30/*'0'*/) * 10 + ((int) date[15] - 0x30/*'0'*/);
  s = ((int) date[17] - 0x30/*'0'*/) * 10 + ((int) date[18] - 0x30/*'0'*/);
  stamp.tm_sec = s; stamp.tm_min = m; stamp.tm_hour = h;
  stamp.tm_mday = day; stamp.tm_mon = month; stamp.tm_year = year;
  stamp.tm_wday = stamp.tm_yday = stamp.tm_isdst = -1;
  t = mktime (&stamp);
  return
    stamp.tm_sec == s && stamp.tm_min == m && stamp.tm_hour == h &&
    stamp.tm_mday == day && stamp.tm_mon == month && stamp.tm_year == year
    ? t
    : -1;
}

/** Complain about invalid Exif DateTime record
 * @param exif_data	the Exif data
 * @param endian	the byte order of the data
 * @param offset	record offset in exif_data
 * @param name		record tag name
 */
static void
exif_date_invalid (const JOCTET FAR* exif_data,
		   enum endianness endian,
		   unsigned long offset,
		   const char* name)
{
  fputs (srcname, stderr);
  fputs (": Exif: invalid ", stderr);
  fputs (name, stderr);
  if (exif_read2 (&exif_data[offset + 2], endian) == 2) {
    unsigned long string_offset, i, len;
    len = exif_read4 (&exif_data[offset + 4], endian);
    if (len > 19) len = 19;
    string_offset = exif_read4 (&exif_data[offset + 8], endian);
    putc ('=', stderr);
    putc ('\'', stderr);
    for (i = 0; i < len; i++) {
      int c = (unsigned char) exif_data[string_offset + i];
      if (c == '\'' || c == '\\')
	putc ('\\', stderr), putc (c, stderr);
      if (isprint (c))
	putc (c, stderr);
      else
	fprintf (stderr, "\\x%02x", c);
    }
    putc ('\'', stderr);
  }
  putc ('\n', stderr);
}

/* Validate string data fields in an Exif record
 * @param exif_data	the Exif data
 * @param endian	the byte order of the data
 * @param offset	record offset in exif_data
 * @param min_offset	start address of data heap
 * @param max_offset	end address of data heap
 * @param ifd1_offset	0 or start address of IFD1
 * @return		zero if the string is valid; 3 if stored within IFD1
 */
static int
exif_string_validate (JOCTET FAR* exif_data,
		      enum endianness endian,
		      unsigned offset,
		      unsigned min_offset,
		      unsigned max_offset,
		      unsigned ifd1_offset)
{
  unsigned long len, ofs;
  if (exif_read2 (&exif_data[offset + 2], endian) != 2)
    return 0;
  len = exif_read4 (&exif_data[offset + 4], endian);
  if (len <= 4)
    return 0; /* stored inline */
  ofs = exif_read4 (&exif_data[offset + 8], endian);
  if (ofs < min_offset) {
    fprintf (stderr, "%s: Exif string offset out of bounds: %lu<%u\n",
	     srcname, ofs, min_offset);
    return 1;
  }
  if (ofs > max_offset - len) {
    fprintf (stderr, "%s: Exif string offset out of bounds: %lu>%lu\n",
	     srcname, ofs, max_offset - len);
    return 1;
  }
  if (exif_data[ofs + len - 1]) {
    fprintf (stderr, "%s: unterminated Exif string at %lu\n",
	     srcname, ofs);
    return 2;
  }
  if (ifd1_offset && ofs > ifd1_offset - len) {
    fprintf (stderr, "%s: string is stored within IFD1\n", srcname);
    return 3;
  }
  return 0;
}

/** Parse the Exif information
 * @param cinfo		decompression structure
 * @param xform		automatic (0) or enum JXFORM_CODE
 * @param timestamp	(input) pointer to timestamp to collect:
 *			0=DateTime, 1=DateTimeOriginal, 2=DateTimeDigitized
 *			(output) the timestamp of the image in the Exif data
 * @param rmifd1	flag: 1=remove IFD1 (containing the thumbnail image),
 *			2=do not touch Exif data, 0=normal processing
 * @param thumbname	Exif thumbnail file name (optional)
 * @return		operation to perform on the image, 0 if not Exif
 */
static enum JXFORM_CODE
exif_parse (struct jpeg_decompress_struct* cinfo,
	    enum JXFORM_CODE xform,
	    time_t* timestamp,
	    int rmifd1,
	    const char* thumbname)
{
  jpeg_saved_marker_ptr marker;

  for (marker = cinfo->marker_list; marker; marker = marker->next) {
    /** byte order of the Exif data */
    enum endianness endian;
    /** current image file directory (IFD) offset */
    unsigned long offset;

    if (marker->marker != (JOCTET) (JPEG_APP0 + 1) ||
	marker->data_length < 14)
      continue; /* Exif data is in an APP1 marker of at least 14 octets */

    if (GETJOCTET (marker->data[0]) != (JOCTET) 0x45 ||
	GETJOCTET (marker->data[1]) != (JOCTET) 0x78 ||
	GETJOCTET (marker->data[2]) != (JOCTET) 0x69 ||
	GETJOCTET (marker->data[3]) != (JOCTET) 0x66 ||
	GETJOCTET (marker->data[4]) != (JOCTET) 0x00 ||
	GETJOCTET (marker->data[5]) != (JOCTET) 0x00)
      continue; /* no Exif header */

    if (GETJOCTET (marker->data[6]) == (JOCTET) 0x4d &&
	GETJOCTET (marker->data[7]) == (JOCTET) 0x4d &&
	GETJOCTET (marker->data[8]) == (JOCTET) 0x00 &&
	GETJOCTET (marker->data[9]) == (JOCTET) 0x2a)
      endian = MSB;
    else if (GETJOCTET (marker->data[6]) == (JOCTET) 0x49 &&
	     GETJOCTET (marker->data[7]) == (JOCTET) 0x49 &&
	     GETJOCTET (marker->data[8]) == (JOCTET) 0x2a &&
	     GETJOCTET (marker->data[9]) == (JOCTET) 0x00)
      endian = LSB;
    else {
      fprintf (stderr, "%s: ignoring Exif marker with byte order "
	       "0x%02x%02x%02x%02x\n",
	       srcname,
	       (unsigned char) GETJOCTET (marker->data[6]),
	       (unsigned char) GETJOCTET (marker->data[7]),
	       (unsigned char) GETJOCTET (marker->data[8]),
	       (unsigned char) GETJOCTET (marker->data[9]));
      continue;
    }

    /* Get first IFD offset (offset to IFD0) */
    offset = exif_read4 (&marker->data[10], endian);
    if (offset > (unsigned long) marker->data_length - (10 + 2)) {
      fprintf (stderr, "%s: Exif IFD0 offset out of bounds: %lu>%lu\n",
	       srcname,
	       offset, (unsigned long) marker->data_length - (10 + 2));
      return 0;
    }
    else {
      /** pointer to the start of the Exif data */
      JOCTET FAR* exif_data = &marker->data[6];
      /** length of Exif data */
      unsigned exif_length = marker->data_length - 6;
      /** Exif SubIFD offset */
      unsigned long subifd_offset;
      /** Exif Interoperability offset */
      unsigned long inop_offset;
      /** Location of the Exif Interoperability Offset */
      unsigned int inop_offset_ptr;
      /** Exif IFD1 offset */
      unsigned long ifd1_offset;
      /** Location of the Exif IFD1 offset */
      unsigned int ifd1_offset_ptr;
      /** Location of the ExifImageWidth */
      unsigned int width_ptr = 0;
      /** Location of the ExifImageHeight */
      unsigned int height_ptr = 0;
      /** Location of the JpegIFByteCount */
      unsigned int bytecount_ptr = 0;
      /** Location of the LSB of the RelatedImageWidth tag */
      unsigned int ri_x_ptr = 0;
      /** Location of the LSB of the RelatedImageHeight tag */
      unsigned int ri_y_ptr = 0;
      /** Location of the LSB of the XResolution tag in IFD0 */
      unsigned int res_x0_ptr = 0;
      /** Location of the LSB of the YResolution tag in IFD0 */
      unsigned int res_y0_ptr = 0;
      /** Location of the LSB of the XResolution tag in IFD1 */
      unsigned int res_x1_ptr = 0;
      /** Location of the LSB of the YResolution tag in IFD1 */
      unsigned int res_y1_ptr = 0;
      /** Location of the LSB of the FocalPlaneXResolution tag */
      unsigned int fres_x_ptr = 0;
      /** Location of the LSB of the FocalPlaneYResolution tag */
      unsigned int fres_y_ptr = 0;
      /** number of entries in the current image file directory */
      unsigned int ifdsize;
      /** end of current image file directory */
      unsigned int ifdend;
      /** thumbnail image size */
      unsigned long thumb_size = 0;
      /** thumbnail image offset */
      unsigned long thumb_offset = 0;
      /** string offset in a string-based IFD entry */
      unsigned long string_offset = 0;
      /** the contents of the DateTime field */
      time_t datetime = -1;
      /** the contents of the DateTimeDigitized field */
      time_t datetimeDigitized = -1;
      /** the contents of the DateTimeOriginal field */
      time_t datetimeOriginal = -1;
      /** offset to DateTime */
      unsigned long datetime_offset = 0;
      /** offset to DateTimeDigitized */
      unsigned long datetimeDigitized_offset = 0;
      /** offset to DateTimeOriginal */
      unsigned long datetimeOriginal_offset = 0;
      /** present orientation (0=unknown, 1..8=known) */
      int orientation = 0;

      /* Get the number of IFD0 entries */
      ifdsize = exif_read2 (&exif_data[offset], endian);
      offset += 2;
      /* Get the IFD1 offset (thumbnail image) */
      ifd1_offset_ptr = ifdend = (unsigned) offset + 12 * ifdsize;
      if (ifdend > exif_length - 4) {
	fprintf (stderr, "%s: Exif IFD0 ends prematurely\n", srcname);
	return 0;
      }
      ifd1_offset = exif_read4 (&exif_data[ifdend], endian);
      if (ifd1_offset) {
	if (ifd1_offset < ifdend + 4) {
	  fprintf (stderr, "%s: Exif IFD1 offset out of bounds: %lu<%lu\n",
		   srcname,
		   ifd1_offset, (unsigned long) ifdend + 4);
	  rmifd1 = 2;
	  ifd1_offset = 0;
	}
	else if (ifd1_offset > exif_length - 2) {
	  fprintf (stderr, "%s: Exif IFD1 offset out of bounds: %lu>%lu\n",
		   srcname,
		   ifd1_offset, (unsigned long) exif_length - 2);
	  rmifd1 = 2;
	  ifd1_offset = 0;
	}
      }
      /* Exif SubIFD and Interoperability offsets */
      subifd_offset = inop_offset = 0;
      inop_offset_ptr = 0;

      /* Search for interesting tags in IFD0 */
      for (; ifdsize--; offset += 12) {
	unsigned int tagnum = exif_read2 (&exif_data[offset], endian);

	if (exif_string_validate (exif_data, endian, offset,
				  ifdend, exif_length,
				  rmifd1 != 2 ? ifd1_offset : 0))
	  rmifd1 = 2;

	switch (tagnum) {
	default:
	  break;
	case 0x0112: /* Orientation */
	  if (orientation) {
	    fprintf (stderr,
		     "%s: Exif: ignoring previous Orientation=%d\n",
		     srcname, orientation);
	    orientation = 0;
	  }
	  if (exif_read2 (&exif_data[offset + 2], endian) != 3 ||
	      exif_read4 (&exif_data[offset + 4], endian) != 1)
	    fprintf (stderr, "%s: Exif: invalid Orientation record\n",
		     srcname);
	  else if ((orientation =
		    exif_read2 (&exif_data[offset + 8], endian)) < 1 ||
		   orientation > 8) {
	    fprintf (stderr, "%s: Exif: invalid Orientation=%d\n",
		     srcname, orientation);
	    orientation = 0;
	  }
	  if (orientation)
	    exif_write (&exif_data[offset], endian, 1/* top, left-hand */);
	  break;
	case 0x011a: /* XResolution */
	  if (res_x0_ptr)
	    fprintf (stderr, "%s: Exif IFD0: ignoring previous XResolution\n",
		     srcname);
	  res_x0_ptr = (unsigned) offset + endian;
	  break;
	case 0x011b: /* YResolution */
	  if (res_y0_ptr)
	    fprintf (stderr, "%s: Exif IFD0: ignoring previous YResolution\n",
		     srcname);
	  res_y0_ptr = (unsigned) offset + endian;
	  break;
	case 0x0132: /* DateTime */
	  if (exif_read2 (&exif_data[offset + 2], endian) != 2 ||
	      exif_read4 (&exif_data[offset + 4], endian) != 20 ||
	      (string_offset = exif_read4 (&exif_data[offset + 8], endian)) <
	      ifdend ||
	      string_offset > (ifd1_offset ? ifd1_offset : exif_length) - 20) {
	  datetime_invalid:
	    exif_date_invalid (exif_data, endian, offset, "DateTime");
	  }
	  else {
	    time_t time = exif_date (exif_data + string_offset);
	    if (time == -1)
	      goto datetime_invalid;
	    if (datetime_offset)
	      fprintf (stderr, "%s: Exif: ignoring previous DateTime='%s'\n",
		       srcname, &exif_data[datetime_offset]);
	    datetime_offset = string_offset;
	    datetime = time;
	  }
	  break;
	case 0x8769: /* Exif SubIFD offset */
	  if (subifd_offset)
	    fprintf (stderr, "%s: "
		     "ignoring previous Exif SubIFD offset=%lu\n",
		     srcname, subifd_offset);
	  if (exif_read2 (&exif_data[offset + 2], endian) != 4 ||
	      exif_read4 (&exif_data[offset + 4], endian) != 1)
	    fprintf (stderr,
		     "%s: invalid format of Exif SubIFD offset\n",
		     srcname);
	  else if ((subifd_offset =
		    exif_read4 (&exif_data[offset + 8], endian)) < ifdend)
	    fprintf (stderr,
		     "%s: Exif SubIFD offset out of bounds: %lu<%u\n",
		     srcname, subifd_offset, ifdend);
	  else if (subifd_offset > exif_length - 2)
	    fprintf (stderr,
		     "%s: Exif SubIFD offset out of bounds: %lu>%u\n",
		     srcname, subifd_offset, exif_length - 2);
	  else if (ifd1_offset && subifd_offset > ifd1_offset - 2)
	    fprintf (stderr, "%s: Exif SubIFD is stored within IFD1\n",
		     srcname);
	  else
	    break;
	  rmifd1 = 2;
	  break;
	}
      }

      if ((offset = subifd_offset)) {
	/* Get the number of SubIFD entries */
	ifdsize = exif_read2 (&exif_data[offset], endian);
	offset += 2;
	ifdend = (unsigned) offset + 12 * ifdsize;
	if (ifdend > (ifd1_offset ? ifd1_offset : exif_length) - 4) {
	  fprintf (stderr, "%s: Exif SubIFD ends prematurely\n", srcname);
	  rmifd1 = 2;
	  ifdsize = 0;
	}
	else if (exif_read4 (&exif_data[ifdend], endian)) {
	  fprintf (stderr, "%s: Exif SubIFD: unexpected 'next' pointer\n",
		   srcname);
	  rmifd1 = 2;
	}

	for (; ifdsize--; offset += 12) {
	  unsigned int tagnum = exif_read2 (&exif_data[offset], endian);

	  if (exif_string_validate (exif_data, endian, offset,
				    ifdend, exif_length,
				    rmifd1 != 2 ? ifd1_offset : 0))
	    rmifd1 = 2;

	  switch (tagnum) {
	  case 0x9003: /* DateTimeOriginal */
	  case 0x9004: /* DateTimeDigitized */
	    if (exif_read2 (&exif_data[offset + 2], endian) != 2 ||
		exif_read4 (&exif_data[offset + 4], endian) != 20 ||
		(string_offset = exif_read4 (&exif_data[offset + 8], endian)) <
		ifdend ||
		string_offset >
		(ifd1_offset ? ifd1_offset : exif_length) - 20) {
	    datetime_invalid2:
	      exif_date_invalid (exif_data, endian, offset,
				 tagnum == 0x9003
				 ? "DateTimeOriginal"
				 : "DateTimeDigitized");
	    }
	    else {
	      time_t time = exif_date (exif_data + string_offset);
	      if (time == -1)
		goto datetime_invalid2;
	      if (tagnum == 0x9003) {
		if (datetimeOriginal_offset)
		  fprintf (stderr, "%s: Exif: ignoring previous"
			   " DateTimeOriginal='%s'\n",
			   srcname, &exif_data[datetimeOriginal_offset]);
	      }
	      else {
		if (datetimeDigitized_offset)
		  fprintf (stderr, "%s: Exif: ignoring previous"
			   " DateTimeDigitized='%s'\n",
			   srcname, &exif_data[datetimeDigitized_offset]);
		datetimeDigitized_offset = string_offset;
		datetimeDigitized = time;
	      }
	    }
	    break;
	  case 0xa002: /* ExifImageWidth */
	  case 0xa003: /* ExifImageHeight */
	    if (exif_read4 (&exif_data[offset + 4], endian) != 1)
	      goto errImageDimension;
	    switch (exif_read2 (&exif_data[offset + 2], endian)) {
	    case 3: /* unsigned short */
	    case 4: /* unsigned long */
	      break;
	    default:
	    errImageDimension:
	      fprintf (stderr, tagnum == 0xa002
		       ? "%s: invalid format of ExifImageWidth\n"
		       : "%s: invalid format of ExifImageHeight\n",
		       srcname);
	      rmifd1 = 2;
	    }
	    if (tagnum == 0xa002) {
	      if (width_ptr)
		fprintf (stderr, "%s: ignoring previous ExifImageWidth\n",
			 srcname);
	      width_ptr = (unsigned) offset;
	    }
	    else {
	      if (height_ptr)
		fprintf (stderr, "%s: ignoring previous ExifImageHeight\n",
			 srcname);
	      height_ptr = (unsigned) offset;
	    }
	    break;
	  case 0xa005: /* ExifInteroperabilityOffset */
	    if (inop_offset_ptr)
	      fprintf (stderr, "%s: "
		       "ignoring previous ExifInteroperabilityOffset=%lu\n",
		       srcname, inop_offset);
	    if (exif_read2 (&exif_data[offset + 2], endian) != 4 ||
		exif_read4 (&exif_data[offset + 4], endian) != 1)
	      fprintf (stderr,
		       "%s: invalid format of ExifInteroperabilityOffset\n",
		       srcname);
	    else if ((inop_offset =
		      exif_read4 (&exif_data[inop_offset_ptr = offset + 8],
				  endian)) < ifdend)
	      fprintf (stderr,
		       "%s: ExifInteroperabilityOffset out of bounds: "
		       "%lu<%u\n",
		       srcname, inop_offset, ifdend);
	    else if (inop_offset > exif_length - 2)
	      fprintf (stderr,
		       "%s: ExifInteroperabilityOffset out of bounds: "
		       "%lu>%u\n",
		       srcname, inop_offset, exif_length - 2);
	    else if (ifd1_offset && inop_offset > ifd1_offset - 2)
	      fprintf (stderr, "%s: Exif InopIFD is stored within IFD1\n",
		       srcname);
	    else
	      break;
	    rmifd1 = 2;
	    break;
	  case 0xa20e: /* FocalPlaneXResolution */
	    if (fres_x_ptr)
	      fprintf (stderr, "%s: Exif SubIFD: ignoring previous"
		       " FocalPlaneXResolution\n", srcname);
	    fres_x_ptr = (unsigned) offset + endian;
	    break;
	  case 0xa20f: /* FocalPlaneYResolution */
	    if (fres_y_ptr)
	      fprintf (stderr, "%s: Exif SubIFD: ignoring previous"
		       " FocalPlaneYResolution\n", srcname);
	    fres_y_ptr = (unsigned) offset + endian;
	    break;
	  }
	}
      }

      if ((offset = inop_offset)) {
	/* Get the number of Interoperability IFD entries */
	ifdsize = exif_read2 (&exif_data[offset], endian);
	offset += 2;
	ifdend = (unsigned) offset + 12 * ifdsize;
	if (ifdend > exif_length - 4) {
	  fprintf (stderr, "%s: Exif InopIFD ends prematurely\n", srcname);
	  rmifd1 = 2;
	  ifdsize = 0;
	}
	else if (exif_read4 (&exif_data[ifdend], endian)) {
	  fprintf (stderr, "%s: Exif InopIFD: unexpected 'next' pointer\n",
		   srcname);
	  rmifd1 = 2;
	}

	for (; ifdsize--; offset += 12) {
	  unsigned int tagnum = exif_read2 (&exif_data[offset], endian);

	  if (exif_string_validate (exif_data, endian, offset,
				    ifdend, exif_length,
				    rmifd1 != 2 ? ifd1_offset : 0))
	    rmifd1 = 2;

	  switch (tagnum) {
	  case 0x1001: /* RelatedImageWidth */
	    if (ri_x_ptr)
	      fprintf (stderr, "%s: ignoring previous RelatedImageWidth\n",
		       srcname);
	    ri_x_ptr = (unsigned) offset + endian;
	    break;
	  case 0x1002: /* RelatedImageHeight */
	    if (ri_y_ptr)
	      fprintf (stderr, "%s: ignoring previous RelatedImageHeight\n",
		       srcname);
	    ri_y_ptr = (unsigned) offset + endian;
	    break;
	  }
	}
      }

      if ((offset = ifd1_offset)) {
	/* Get the number of IFD1 entries */
	ifdsize = exif_read2 (&exif_data[offset], endian);
	offset += 2;
	ifdend = offset + 12 * ifdsize;
	if (ifdend > exif_length - 4) {
	  fprintf (stderr, "%s: Exif IFD1 ends prematurely\n", srcname);
	  ifdsize = 0;
	  rmifd1 = 2;
	}
	/* Get the next IFD offset (should be zero) */
	if (exif_read4 (&exif_data[ifdend], endian)) {
	  fprintf (stderr, "%s: nonnull 'next' pointer in Exif IFD1\n",
		   srcname);
	  rmifd1 = 2;
	}
	for (; ifdsize--; offset += 12) {
	  unsigned int tagnum = exif_read2 (&exif_data[offset], endian);

	  if (exif_string_validate (exif_data, endian, offset,
				    ifdend, exif_length, 0))
	    rmifd1 = 2;

	  switch (tagnum) {
	  case 0x011a: /* XResolution */
	    if (res_x1_ptr)
	      fprintf (stderr, "%s: Exif IFD1: ignoring previous"
		       " XResolution\n", srcname);
	    res_x1_ptr = (unsigned) offset + endian;
	    break;
	  case 0x011b: /* YResolution */
	    if (res_y1_ptr)
	      fprintf (stderr, "%s: Exif IFD1: ignoring previous"
		       " YResolution\n", srcname);
	    res_y1_ptr = (unsigned) offset + endian;
	    break;
	  case 0x201: /* JpegIFOffset */
	    if (thumb_offset)
	      fprintf (stderr, "%s: Exif: ignoring previous JpegIFOffset\n",
		       srcname);
	    if (exif_read2 (&exif_data[offset + 2], endian) != 4 ||
		exif_read4 (&exif_data[offset + 4], endian) != 1)
	      fprintf (stderr, "%s: invalid format of Exif JpegIFOffset\n",
		       srcname);
	    else if ((thumb_offset =
		      exif_read4 (&exif_data[offset + 8], endian)) < ifdend)
	      fprintf (stderr,
		       "%s: Exif JpegIFOffset out of bounds: %lu<%u\n",
		       srcname, thumb_offset, ifdend);
	    else if (thumb_offset + thumb_size > exif_length + 1)
	      fprintf (stderr,
		       "%s: Exif JpegIFOffset out of bounds: %lu>%u\n",
		       srcname, thumb_offset + thumb_size, exif_length);
	    /* "exif_length + 1" tolerates an off-by-1 error
	       of Sony DCR-PC110E */
	    else
	      break;
	    rmifd1 = 2;
	    break;
	  case 0x202: /* JpegIFByteCount */
	    if (exif_read2 (&exif_data[offset + 2], endian) != 4 ||
		exif_read4 (&exif_data[offset + 4], endian) != 1)
	      fprintf (stderr, "%s: invalid format of Exif JpegIFByteCount\n",
		       srcname);
	    else if ((thumb_size =
		      exif_read4 (&exif_data[offset + 8], endian)) &&
		     thumb_offset + thumb_size > exif_length + 8)
	      fprintf (stderr, "%s: Exif JpegIFByteCount out of bounds:"
		       " %lu+%lu>%u\n",
		       srcname, thumb_offset, thumb_size, exif_length);
	    /* "exif_length + 8" tolerates an off-by-1 error
	       of Sony DCR-PC110E and an off-by-8 error of Canon EOS 350D
	       firmware 1.0.2, May 2005. */
	    else {
	      if (bytecount_ptr)
		fprintf (stderr,
			 "%s: Exif: ignoring previous JpegIFByteCount\n",
			 srcname);
	      bytecount_ptr = offset;
	      break;
	    }
	    rmifd1 = 2;
	    break;
	  }
	}
      }

      if (thumb_offset && thumb_size && thumbname) {
	FILE* f = fopen (thumbname, "wb");
	if (!f) {
	  (void) fputs (thumbname, stderr), (void) fflush (stderr);
	  perror (": open (writing)");
	  thumbname = 0;
	}
	else {
	  if (thumb_offset + thumb_size > exif_length)
	    thumb_size = exif_length - thumb_offset;
	  if (exif_data[thumb_offset + thumb_size - 1] == (JOCTET) 0)
	    thumb_size--; /* some cameras add a terminating NUL character */
	  /* output the thumbnail */
	  if (thumb_size !=
	      fwrite (exif_data + thumb_offset, 1, thumb_size, f)) {
	    (void) fputs (thumbname, stderr), (void) fflush (stderr);
	    perror (": write");
	    thumbname = 0;
	  }
	  if (fclose (f) && thumbname) {
	    (void) fputs (thumbname, stderr), (void) fflush (stderr);
	    perror (": close");
	    thumbname = 0;
	  }
	}
      }

      if (rmifd1 == 2)
	fprintf (stderr, "%s: Exif block inconsistent: will not touch it\n",
		 srcname);
      else if (rmifd1 == 1 && ifd1_offset) {
	/* remove the IFD1 entry (which contains the thumbnail image) */
	exif_data[ifd1_offset_ptr] =
	  exif_data[ifd1_offset_ptr + 1] = 
	  exif_data[ifd1_offset_ptr + 2] = 
	  exif_data[ifd1_offset_ptr + 3] = (JOCTET) 0;
	exif_length = (unsigned) ifd1_offset;
	marker->data_length = exif_length + 6;
      }

      if (timestamp) {
	switch (*timestamp) {
	default: *timestamp = datetime; break;
	case 1: *timestamp = datetimeOriginal; break;
	case 2: *timestamp = datetimeDigitized; break;
	}

	if (thumb_offset && thumb_size && thumbname)
	  touchfile (thumbname, *timestamp);
      }

      if (xform < 1 || xform > 8)
	if (!(xform = orientation))
	  xform = JXFORM_NONE;

      /* adjust ExifImageWidth and ExifImageHeight */
      if (rmifd1 != 2 && width_ptr && height_ptr) {
	unsigned height = cinfo->image_height, width = cinfo->image_width;
	switch (xform) {
	case JXFORM_TRANSPOSE:
	case JXFORM_TRANSVERSE:
	case JXFORM_ROT_90:
	case JXFORM_ROT_270:
	  width = cinfo->image_height, height = cinfo->image_width;
	  break;
	case JXFORM_NONE:
	case JXFORM_FLIP_H:
	case JXFORM_FLIP_V:
	case JXFORM_ROT_180:
	  break;
	}
	exif_write (&exif_data[width_ptr], endian, width);
	exif_write (&exif_data[height_ptr], endian, height);
      }
      /* flip the x/y parameters if needed */
      switch (xform) {
      case JXFORM_TRANSPOSE:
      case JXFORM_TRANSVERSE:
      case JXFORM_ROT_90:
      case JXFORM_ROT_270:
	if (rmifd1 == 2) break;
	if (ri_x_ptr) exif_data[ri_x_ptr] = (JOCTET) 2;
	if (ri_y_ptr) exif_data[ri_y_ptr] = (JOCTET) 1;
	if (res_x0_ptr) exif_data[res_x0_ptr] = (JOCTET) 0x1b;
	if (res_y0_ptr) exif_data[res_y0_ptr] = (JOCTET) 0x1a;
	if (res_x1_ptr) exif_data[res_x1_ptr] = (JOCTET) 0x1b;
	if (res_y1_ptr) exif_data[res_y1_ptr] = (JOCTET) 0x1a;
	if (fres_x_ptr) exif_data[fres_x_ptr] = (JOCTET) 0x0f;
	if (fres_y_ptr) exif_data[fres_y_ptr] = (JOCTET) 0x0e;
	break;
      case JXFORM_NONE:
      case JXFORM_FLIP_H:
      case JXFORM_FLIP_V:
      case JXFORM_ROT_180:
	break;
      }

      /* transform the thumbnail image */
      if (thumb_offset && thumb_size && !rmifd1) {
	thumb_transform (xform, exif_data + thumb_offset, thumb_size,
			 65533 - thumb_offset);
	if (thumb_length) {
	  /* if the transformation succeeded, adjust the thumbnail length
	   * and remove the original thumbnail image
	   * (the new one will be written by jcopy_markers_execute)
	   */
	  exif_length = thumb_offset;
	  marker->data_length = exif_length + 6;
	  switch (endian) {
	  case LSB:
	    exif_data[bytecount_ptr + 8] = (JOCTET) thumb_length;
	    exif_data[bytecount_ptr + 9] = (JOCTET) (thumb_length >> 8);
	    break;
	  case MSB:
	    exif_data[bytecount_ptr + 11] = (JOCTET) thumb_length;
	    exif_data[bytecount_ptr + 10] = (JOCTET) (thumb_length >> 8);
	    break;
	  }
	}
      }
      return xform;
    }
  }

  return 0;
}

/**
 * Construct a file name
 * @param base		the base file name
 * @param baselen	length of the base file name
 * @param dir		the directory name
 * @param dirlen	length of the directory name
 * @return		the constructed file name, or 0 on error
 */
static char*
createfilename (const char* base,
		unsigned baselen,
		const char* dir,
		unsigned dirlen)
{
  const char* s;
  char* fnam;
  for (s = base + baselen; s > base; s--)
    if (*s == '/'
#if defined WIN32 || defined __WIN32
	|| *s == '\\' || *s == ':'
#endif
	) { s++; break; }
  baselen -= s - base;
  fnam = malloc (dirlen + baselen + 2);
  if (fnam) {
    memcpy (fnam, dir, dirlen);
    fnam[dirlen] = '/';
    memcpy (fnam + dirlen + 1, s, baselen + 1);
  }
  return fnam;
}

/** Transform the image
 * @param srcinfo	image decompression context
 * @param dstinfo	image compression context
 * @param thumbname_	thumbnail file name (optional)
 * @param dstname_	destination image file name (NULL = stdin/stdout)
 * @param xform_	image transformation
 * @param rmifd1_	flag: 1=remove Exif image file directory 1 (thumbnail),
 *			2=do not touch Exif data, 0=normal processing
 * @param datetime	(output) timestamp to collect from Exif block
 * @param progressive	flag: use progressive coding
 * @param optimize	flag: compute optimal Huffman coding tables
 * @return		0=ok, 1=read error, 2=write error
 */
static int
jpeg_transform (struct jpeg_decompress_struct* srcinfo,
		struct jpeg_compress_struct* dstinfo,
		const char* thumbname_,
		const char* dstname_,
		enum JXFORM_CODE xform_,
		int rmifd1_,
		time_t* datetime,
		int progressive,
		int optimize)
{
  /** source file handle */
  FILE* src;
  /** destination file handle */
  FILE* dst;
  /** image transformation information */
  struct jpeg_transform_info info;
  /** flag: is this a JFIF image (not Exif) */
  int jfif = 0;
  /** coefficient arrays of source image */
  struct jvirt_barray_control** src_coef_arrays;
  /** coefficient arrays of destination image */
  struct jvirt_barray_control** dst_coef_arrays;
  /** thumbnail file name (optional) */
  const char* thumbname = thumbname_;
  /** destination image file name (NULL = stdin/stdout) */
  const char* dstname = dstname_;
  /** image transformation */
  enum JXFORM_CODE xform = xform_;
  /** flag: 1=remove Exif image file directory 1 (thumbnail) */
  int rmifd1 = rmifd1_;

  /* clear possible previous errors */
  errno = 0;
  src = dst = 0;

  /* set up jump context for input */
  if (setjmp (((struct exif_error_mgr*) srcinfo->err)->setjmp_buffer)) {
    if (errno)
      perror (dstname ? srcname : "(standard input)");
    else
      exif_output_message ((j_common_ptr) srcinfo);
    jpeg_abort_decompress (srcinfo);
    jpeg_abort_compress (dstinfo);
    if (src) (void) fclose (src);
    if (dst) (void) fclose (dst);
    return 1; /* ignore unreadable or erroneous input files */
  }

  /* set up jump context for output */
  if (setjmp (((struct exif_error_mgr*) dstinfo->err)->setjmp_buffer)) {
    if (errno)
      perror (dstname ? dstname : "(standard output)");
    else
      exif_output_message ((j_common_ptr) dstinfo);
    jpeg_abort_decompress (srcinfo);
    jpeg_abort_compress (dstinfo);
    if (src) (void) fclose (src);
    if (dst) (void) fclose (dst);
    return 2; /* abort on write error */
  }

  if (dstname && !(src = fopen (srcname, "rb"))) {
    (void) fputs (srcname, stderr), (void) fflush (stderr);
    perror (": open (reading)");
    return 1;
  }
  jpeg_stdio_src (srcinfo, dstname ? src : stdin);
  jcopy_markers_setup (srcinfo);
  (void) jpeg_read_header (srcinfo, TRUE);

  info.transform = exif_parse (srcinfo, xform, datetime, rmifd1, thumbname);
  jfif = !info.transform;
  if (jfif)
    info.transform = xform ? xform : JXFORM_NONE;

  jtransform_request_workspace (srcinfo, &info);

  /* Read source file as DCT coefficients */
  src_coef_arrays = jpeg_read_coefficients (srcinfo);
  /* Close the input file */
  if (src && fclose (src)) {
    (void) fputs (srcname, stderr), (void) fflush (stderr);
    perror (": close");
    /* ignore error on closing input file */
  }
  src = 0;
  /* Initialize destination compression parameters from source values */
  jpeg_copy_critical_parameters (srcinfo, dstinfo);

  dst_coef_arrays = jtransform_adjust_parameters (dstinfo,
						  src_coef_arrays,
						  &info);
  if (dstname) {
    int fd;
    (void) unlink (dstname);
    fd = open (dstname,
#if defined WIN32 || defined __WIN32
	       O_WRONLY | O_CREAT | O_EXCL | O_BINARY, 0666
#else
	       O_WRONLY | O_CREAT | O_EXCL, 0444
#endif
	       );
    if (fd < 0 || !(dst = fdopen (fd, "wb"))) {
      (void) fputs (dstname, stderr), (void) fflush (stderr);
      perror (": open (writing)");
      return 2;
    }
  }
  jpeg_stdio_dest (dstinfo, dstname ? dst : stdout);
  dstinfo->optimize_coding = optimize;
  dstinfo->write_JFIF_header = jfif;
  if (progressive)
    jpeg_simple_progression (dstinfo);
  jpeg_write_coefficients (dstinfo, dst_coef_arrays);
  jcopy_markers_execute (srcinfo, dstinfo, thumb_buf, thumb_length);
  jtransform_execute_transform (srcinfo, dstinfo, src_coef_arrays, &info);
  (void) jpeg_finish_compress (dstinfo);
  (void) jpeg_finish_decompress (srcinfo);
  if (dst && fclose (dst)) {
    (void) fputs (dstname, stderr), (void) fflush (stderr);
    perror (": close");
    return 2;
  }
  return 0;
}

int
main (int argc, char **argv)
{
  /** libjpeg decompression context */
  struct jpeg_decompress_struct srcinfo;
  /** libjpeg compression context */
  struct jpeg_compress_struct dstinfo;
  /** libjpeg error handlers */
  struct exif_error_mgr jsrcerr, jdsterr;
  /** exit status */
  int status = 0;

  /** flag: remove Exif IFD1 */
  int rmifd1 = 1;
  /** image transformation (0=according to the Orientation tag) */
  enum JXFORM_CODE transformation = 0;
  /** time stamp for output files */
  enum { tdt, tdtDigitized, tdtOriginal, tdtNone } tOutput = tdt;
  /** flag: disable the processing of options */
  int ignopt = 0;
  /** flag: use the progressive JPEG process */
  int progressive = 1;
  /** flag: compute optimal Huffman coding tables for the image */
  int optimize = 1;
  /** directory for modified image files (0=replace the original files) */
  char* imagedir = 0;
  /** directory for Exif thumbnail files (0=output disabled) */
  char* thumbdir = 0;

  srcinfo.err = jpeg_std_error (&jsrcerr.pub);
  jsrcerr.pub.addon_message_table = exif_message_table;
  jsrcerr.pub.first_addon_message = EXIF_FIRSTCODE;
  jsrcerr.pub.last_addon_message = EXIF_LASTCODE;
  dstinfo.err = jpeg_std_error (&jdsterr.pub);

  jpeg_create_decompress (&srcinfo);
  jpeg_create_compress (&dstinfo);

  jsrcerr.pub.error_exit = jdsterr.pub.error_exit = exif_error_exit;
  jsrcerr.pub.output_message = jdsterr.pub.output_message =
    exif_output_message;

  for (argv++; argc > 1; argv++, argc--) {
    time_t datetime;
    const char* opts = ignopt || **argv != '-' ? 0 : *argv;

    if (opts && !strcmp (opts, "--")) {
      argv++;
      argc--;
      ignopt = 1;
      continue;
    }

    while (opts && *++opts) { /* process all flags */
      switch (*opts) {
      default:
	(void) putc ('-', stderr);
	(void) putc (*opts, stderr);
	(void) fputs (": unknown option (use -h for help)\n", stderr);
	status = 1;
	goto done;
      case '?':
      case 'h':
	(void) fputs
	  ("exifiron 1.2.3\n"
	   "Orientation correction and shrinking of Exif files\n"
	   "Usage: exifiron [options] file ...\n"
	   "Options:\n"
	   "-h"
	   "\tusage\n"
	   "-0"
	   "\tautomatic orientation correction (default)\n"
	   "-1,-u"
	   "\tno transformation\n"
	   "-2,-f"
	   "\thorizontal flip\n"
	   "-3,-U"
	   "\tupside down\n"
	   "-4,-F"
	   "\tvertical flip\n"
	   "-5,-t"
	   "\ttranspose\n"
	   "-6,-R"
	   "\trotate right\n"
	   "-7,-T"
	   "\ttransverse\n"
	   "-8,-r"
	   "\trotate left\n"
	   "-e"
	   "\tremove thumbnail (default)\n"
	   "-E"
	   "\tpreserve thumbnail\n"
	   "-d,-D,-O"
	   "\tstamp with DateTime{,Digitized,Original} (default=-d)\n"
	   "-o"
	   "\tstamp with current time\n"
	   "-b"
	   "\tbaseline JPEG\n"
	   "-p"
	   "\tprogressive JPEG (default)\n", stderr);
	(void) fputs
	  ("-n"
	   "\tdo not compute optimal Huffman coding for the image\n"
	   "-s dir\n"
	   "\tsave the image files in dir instead of replacing originals\n"
	   "-S dir\n"
	   "\tsave the Exif thumbnails in dir\n", stderr);
	goto done;

      case '0': transformation = 0; break;
      case '1': case 'u': transformation = JXFORM_NONE; break;
      case '2': case 'f': transformation = JXFORM_FLIP_H; break;
      case '3': case 'U': transformation = JXFORM_ROT_180; break;
      case '4': case 'F': transformation = JXFORM_FLIP_V; break;
      case '5': case 't': transformation = JXFORM_TRANSPOSE; break;
      case '6': case 'R': transformation = JXFORM_ROT_90; break;
      case '7': case 'T': transformation = JXFORM_TRANSVERSE; break;
      case '8': case 'r': transformation = JXFORM_ROT_270; break;

      case 'e': rmifd1 = 1; break;
      case 'E': rmifd1 = 0; break;

      case 'd': tOutput = tdt; break;
      case 'D': tOutput = tdtDigitized; break;
      case 'O': tOutput = tdtOriginal; break;
      case 'o': tOutput = tdtNone; break;

      case 'b': progressive = 0; break;
      case 'p': progressive = 1; break;

      case 'n': optimize = 0; break;

      case 's': case 'S':
	if (argc <= 2) {
	  putc ('-', stderr), putc (*opts, stderr);
	  (void) fputs (": missing argument (use -h for help)\n", stderr);
	  status = 1;
	  goto done;
	}
	argc--;
	if (!**++argv) {
	  /* empty string argument => restore default behaviour */
	  if (*opts == 's')
	    imagedir = 0;
	  else
	    thumbdir = 0;
	}
	else {
	  unsigned namelen = strlen (*argv);
#if defined WIN32 || defined __WIN32
	  char* s = *argv;
	  while (*s) if (*s++ == '\\') s[-1] = '/';
#endif
	  if (*opts == 's')
	    imagedir = *argv;
	  else
	    thumbdir = *argv;
	  if (mkdir (*argv
#if !defined WIN32 && !defined __WIN32
		     , 0777
#endif
		     ) && errno != EEXIST) {
	    (void) fputs (*argv, stderr), (void) fflush (stderr);
	    perror (": mkdir");
	  quit:
	    status = 2;
	    goto done;
	  }
	  else {
	    struct stat statbuf;
	    char* s = malloc (namelen + 3);
	    if (!s) {
	      putc ('-', stderr), putc (*opts, stderr);
	      (void) fputs (": out of memory\n", stderr);
	      goto quit;
	    }
	    memcpy (s, *argv, namelen);
	    memcpy (s + namelen, "/.", 3);
	    if (stat (s, &statbuf)) {
	      (void) fputs (s, stderr), (void) fflush (stderr);
	      perror (": stat");
	      free (s);
	      goto quit;
	    }
	    free (s);
	  }
	}
	break;
      }
    }

    if (opts && opts != *argv + 1)
      continue; /* get next parameters */

    /* if opts != 0 at this point, it denotes the file name '-' */
    srcname = *argv;
    {
      /** destination file name */
      char* dstname;
      /** thumbnail file name */
      char* thumbname;
      /* allocate the destination file name */
      unsigned len = strlen (*argv);

      if (opts) /* file name '-' */
	dstname = thumbname = 0;
      else {
	thumbname = thumbdir
	  ? createfilename (*argv, len, thumbdir, strlen (thumbdir))
	  : 0;
	if (!(dstname = imagedir
	      ? createfilename (*argv, len, imagedir, strlen (imagedir))
	      : malloc (len + 1))) {
	  (void) fputs (*argv, stderr);
	  (void) fputs (": out of memory\n", stderr);
	  goto quit2;
	}
	if (!imagedir) {
	  memcpy (dstname, *argv, len + 1);
	  dstname[len - 1] = '0'; /* change the last character to '0' */
	}
      }
      datetime = tOutput;
      switch (jpeg_transform (&srcinfo, &dstinfo, thumbname, dstname,
			      transformation, rmifd1,
			      tOutput == tdtNone ? 0 : &datetime,
			      progressive, optimize)) {
      case 0:
	break;
      case 1:
	if (dstname) free (dstname);
	if (thumbname) free (thumbname);
	continue;
      default:
      quit2:
	status = 2;
	if (dstname) free (dstname);
	if (thumbname) free (thumbname);
	goto done;
      }
      if (dstname) {
	if (datetime != (time_t) tOutput)
	  touchfile (dstname, datetime);
	if (!imagedir && rename_file (dstname, *argv))
	  goto quit2;
	free (dstname);
      }
      if (thumbname) free (thumbname);
    }
  }

 done:
  jpeg_destroy_compress (&dstinfo);
  jpeg_destroy_decompress (&srcinfo);
  return status;
}
